unit embeddedapi;

interface

{
  The Chromium browser uses multiplie processes
  - The master "browser" process.
    This is the where the application main form is shown.
  - A "render" process per browser HTML window (frame).
    This is where the HTML is displayed and where the javascript runs.

  The 2 processes communicate through an message API.

  The routines in this unit run in the 'render' process.
  Anything that must be displayed or requested from the user,
  should be sent to the browser process using the message API.

}

uses
  Sysutils, Classes, uCEFChromium, uCEFWindowParent, uCEFInterfaces, uCEFApplication, uCEFTypes, uCEFConstants,
  uCEFv8Value, uCEFv8Accessor, uCEFv8Handler;

Const
  // Messages between renderer/browser
  SMsgSetTrayIconVisibility = 'SetTrayIconVisibility';
  SMsgAddTrayMenuItem = 'addTrayMenuItem';
  SMsgRemoveTrayMenuItem = 'removeTrayMenuItem';
  SMsgExecuteTrayMenuClick = 'ExecuteTrayMenuItemClick';
  SMsgDisplayLogMessage = 'DisplayLogMessage';

  // tray Icon object structure as seen in browser
  SPropVisible = 'visible';
  SFuncAddMenuItem = 'addMenuItem';
  SFuncRemoveMenuItem = 'removeMenuItem';

  // We hook in to console.log, so we can display it in the main form :)
  SFuncDoLog = 'appLog';
  STrayIcon = 'trayIcon';

Type
  (*
    // Tray Icon object.
    {
      visible : boolean;
      Function addMenuItem(aCaption : String; ACallBack : Function);
    }
  *)

  TTrayIconAccessor = class(TCefV8AccessorOwn)
  private
    FVisible : Boolean;
    procedure SendBrowserShowHideTrayIcon(aValue: Boolean);
  protected
    function Get(const name: ustring; const obj: ICefv8Value; var retval : ICefv8Value; var exception: ustring): Boolean; override;
    function Set_(const name: ustring; const obj, value: ICefv8Value; var exception: ustring): Boolean; override;
  end;

  {
    This is a proxy of the tray menu running in the browser.
    The proxy lives in the renderer proces;
  }

  TTrayMenuItem = Class(TCollectionItem)
  private
    FHandler: ICefv8Value;
    FCaption: uString;
    FContext: ICefv8Context;
  Public
    Property Caption : uString Read FCaption Write FCaption;
    Property Handler : ICefv8Value Read FHandler Write FHandler;
    Property Context : ICefv8Context Read FContext Write FContext;
  End;

  { TTrayMenu }

  TTrayMenu = Class(TCollection)
  private
    function GetM(AIndex: Integer): TTrayMenuItem;
  Public
    Function AddMenu(Const aCaption : uString) : TTrayMenuItem;
    Function RemoveMenu(Const aID : Integer): Boolean;
    Property Menus[AIndex : Integer] : TTrayMenuItem Read GetM;
  End;

  // Object methods handler for the tray icon object

  { TTrayIconHandler }

  TTrayIconHandler = class(TCefv8HandlerOwn)
  Private
    Class Var FMenu : TTrayMenu;
    function AddMenuItem(aCaption: uString; AHandler: ICefv8Value; AContext : ICefV8Context): Integer;
    function RemoveMenuItem(aID: Integer; AContext: ICefV8Context): Integer;
    class procedure CheckMenu; static;
  protected
    class procedure CallTrayMenuCallBack(aID : integer); static;
    function Execute(const name: ustring; const obj: ICefv8Value; const arguments: TCefv8ValueArray; var retval: ICefv8Value; var exception: ustring): Boolean; override;
  end;

  // handler for the window.appLog function

  TLogHandler = class(TCefv8HandlerOwn)
    function Execute(const name: ustring; const obj: ICefv8Value; const arguments: TCefv8ValueArray; var retval: ICefv8Value; var exception: ustring): Boolean; override;
  end;

// The following callbacks are used in the main project source file.

// Here we instantiate the tray icon.
// We also create window.appLog here. It is used in the extension.
procedure HandleNewContext(const browser: ICefBrowser; const frame: ICefFrame; const context: ICefv8Context);

// This is the entry point when the browser process sends a message to the render process.
procedure HandleProcessMessage(const browser       : ICefBrowser;
                               const frame: ICefFrame;
                               sourceProcess : TCefProcessId;
                               const aMessage      : ICefProcessMessage;
                               var   aHandled      : boolean);

// If the browser has an unhandled exception, we catch it and send it to the browser proces for display.
procedure HandleJSException(const browser: ICefBrowser; const frame: ICefFrame; const context: ICefv8Context; const exception: ICefV8Exception; const stackTrace: ICefV8StackTrace);

// The console log override is done as early as possible using an 'extension' : a small piece of Javascript that is executed before all else.
procedure HandleWebkitInitialized;

// Utility function to convert an ICefv8Value to a string.
// Calls itself recursively for arrays and objects.
Function CefValToString(A : ICefv8Value) : uString;

implementation

uses uCEFMiscFunctions, uCEFv8Context, uCEFProcessMessage;

Procedure SendProcessMessage(const aProcessID : TCefProcessID; const Msg :ICefProcessMessage);

begin
  TCefv8ContextRef.Current.Browser.MainFrame.SendProcessMessage(aProcessID,Msg);
end;


Function CefValToString(A : ICefv8Value) : uString;

var
  i,j : Integer;
  L : TStrings;

begin
  if A.IsString then
    Result:=A.GetStringValue
  else if A.IsBool then
    Result:=UTF8decode(BoolToStr(A.GetBoolValue))
  else if A.IsInt then
    Result:=UTF8decode(IntToStr(A.GetIntValue))
  else if A.IsDouble then
    Result:=UTF8decode(FloatToStr(A.GetDoubleValue))
  else if A.IsNull then
    Result:='null'
  else if A.IsArray then
    begin
    Result:='[';
    for J:=0 to A.GetArrayLength-1 do
      begin
      if J>0 then
        Result:=Result+', ';
      Result:=Result+CefValToString(A.GetValueByIndex(J));
      end;
    Result:=Result+']';
    end
  else if A.IsObject then
    begin
    Result:='{';
    L:=TstringList.Create;
    try
    A.GetKeys(L);
    For J:=0 to L.Count-1 do
      begin
      if J>0 then
        Result:=Result+', ';
      Result:=Result+'"'+UTF8decode(L[i])+'": '+CefValToString(A.GetValueByKey(L[i]));
      end;
    finally
      l.Free;
    end;
    Result:=Result+']';
    end
  else
    Result:=Result+'[Cannot display this value]';
end;


procedure HandleWebkitInitialized;

var
  HookConsole : string;

begin
  HookConsole:='(function() {' +
               '  var oldlog;'+
               '  if (!console) console = {}; '+
               '  if (console.log) oldlog=console.log; '+
               '  console.log = function() {' +
               '    if (oldlog) oldlog.apply(window,arguments); '+
               '    window.'+SFuncDoLog+'.apply(window,arguments); '+
               '  };' +
               '})();';
  CefRegisterExtension('v8/hookconsole', HookConsole,Nil );
end;

procedure HandleProcessMessage(const browser: ICefBrowser;
                               const frame: ICefFrame;
                               sourceProcess: TCefProcessId;
                               const aMessage: ICefProcessMessage;
                               var   aHandled: boolean);
var
  aID : Integer;

begin
  if (aMessage.name=SMsgExecuteTrayMenuClick)
      and (aMessage.ArgumentList.GetSize>0) then
    begin
    aID:=aMessage.ArgumentList.GetInt(0);
    TTrayIconHandler.CallTrayMenuCallBack(aID);
    aHandled:=true;
    end;
end;

procedure TTrayIconAccessor.SendBrowserShowHideTrayIcon(aValue : Boolean);

var
  msg: ICefProcessMessage;

begin
  msg := TCefProcessMessageRef.New(SMsgSetTrayIconVisibility);
  msg.ArgumentList.SetBool(0,aValue);
  SendProcessMessage(PID_BROWSER, msg);
end;

function TTrayIconAccessor.Get(const name      : ustring;
                           const obj       : ICefv8Value;
                           var   retval    : ICefv8Value;
                           var   exception : ustring): Boolean;
begin
  Result:=(name=SPropVisible);
  if Result then
    retval := TCefv8ValueRef.NewBool(FVisible);
end;

function TTrayIconAccessor.Set_(const name      : ustring;
                           const obj       : ICefv8Value;
                           const value     : ICefv8Value;
                           var   exception : ustring): Boolean;
begin
  Result:=(name=SPropVisible);
  if Result then
    begin
    if value.IsBool then
      begin
      FVisible:=value.GetBoolValue;
      SendBrowserShowHideTrayIcon(FVisible);
      end
    else
      exception := 'Invalid value type, expect boolean';
    end
end;


Function CreateTrayIconObject : ICefv8Value;

Var
  TempAccessor : ICefV8Accessor;
  TempHandler : ICefv8Handler;
  Ref : ICefv8Value;
begin
  TempAccessor := TTrayIconAccessor.Create;
  TempHandler:= TTrayIconHandler.Create;
  // Create object
  Result:=TCefv8ValueRef.NewObject(TempAccessor,nil);
  // Add visible property
  if not Result.SetValueByKey(SPropVisible, TCefv8ValueRef.NewBool(False),V8_PROPERTY_ATTRIBUTE_NONE) then
    Raise Exception.Create('Could not set visible attribute');
  Result.SetValueByAccessor(SPropVisible, V8_ACCESS_CONTROL_DEFAULT,  V8_PROPERTY_ATTRIBUTE_NONE);
  // Add AddMenuItem method
  Ref:=TCefv8ValueRef.NewFunction(SFuncAddMenuItem,TempHandler);
    if not Result.SetValueByKey(SFuncAddMenuItem,Ref,V8_PROPERTY_ATTRIBUTE_NONE) then
    Raise Exception.Create('Could not add addMenuItem function');
  // Add RemoveMenuItem method
  Ref:=TCefv8ValueRef.NewFunction(SFuncRemoveMenuItem,TempHandler);
  if not Result.SetValueByKey(SFuncRemoveMenuItem,Ref,V8_PROPERTY_ATTRIBUTE_NONE) then
    Raise Exception.Create('Could not add addMenuItem function');
end;

Function CreateLoggerObject : ICefv8Value;

begin
  Result:=TCefv8ValueRef.NewFunction(SFuncDoLog,TLogHandler.Create);
end;

procedure HandleNewContext(const browser: ICefBrowser; const frame: ICefFrame; const context: ICefv8Context);

Var
  logger,
  trayIcon : ICefv8Value;

begin
  trayIcon:=CreateTrayIconObject;
  logger:=CreateLoggerObject;
  if not context.Global.SetValueByKey(STrayIcon, trayIcon, V8_PROPERTY_ATTRIBUTE_NONE) then
    Raise Exception.Create('Could not set global tray icon object');
  if not context.Global.SetValueByKey(SFuncDoLog, logger, V8_PROPERTY_ATTRIBUTE_NONE) then
    Raise Exception.Create('Could not set global log function');
end;

{ TTrayIconHandler }
class procedure TTrayIconHandler.CallTrayMenuCallBack(aID: integer);

Var
  M : TTrayMenuItem;
  arguments: TCefv8ValueArray;

begin
  If not Assigned(FMenu)then exit;
  M:= TTrayMenuItem(FMenu.FindItemID(aID));
  if M=Nil then exit;
    begin
    SetLength(arguments,1);
    Arguments[0]:=TCefv8ValueRef.NewInt(aID);
    M.Handler.ExecuteFunctionWithContext(M.Context, nil, arguments);
    end;
end;

Class Procedure TTrayIconHandler.CheckMenu;

begin
  If (FMenu=Nil) then
    FMenu:=TTrayMenu.Create(TTrayMenuItem);
end;

function TTrayIconHandler.AddMenuItem(aCaption : uString; AHandler : ICefv8Value; AContext : ICefV8Context) : Integer;

Var
  M : TTrayMenuItem;
  msg: ICefProcessMessage;

begin
  CheckMenu;
  M:=FMenu.AddMenu(aCaption);
  M.Handler:=aHandler;
  M.Context:=aContext;
  Result:=M.ID;
  // Send message to browser process.
  msg := TCefProcessMessageRef.New(SMsgAddTrayMenuItem);
  msg.ArgumentList.SetSize(2);
  msg.ArgumentList.SetInt(0,M.ID);
  msg.ArgumentList.SetString(1,M.Caption);
  SendProcessMessage(PID_BROWSER, msg);
end;

function TTrayIconHandler.RemoveMenuItem(aID : Integer; AContext : ICefV8Context) : Integer;

Var
  msg: ICefProcessMessage;

begin
  CheckMenu;
  FMenu.RemoveMenu(aID);
  Result:=0;
  // Send message to browser process.
  msg := TCefProcessMessageRef.New(SMsgRemoveTrayMenuItem);
  msg.ArgumentList.SetSize(1);
  msg.ArgumentList.SetInt(0,aID);
  SendProcessMessage(PID_BROWSER, msg);
end;


procedure HandleJSException(const browser: ICefBrowser; const frame: ICefFrame; const context: ICefv8Context; const exception: ICefV8Exception; const stackTrace: ICefV8StackTrace);

Var
  msg: ICefProcessMessage;

begin
  msg := TCefProcessMessageRef.New(SMsgDisplayLogMessage);
  msg.ArgumentList.SetString(0, 'Exception : '+Exception.Message);
  Browser.MainFrame.SendProcessMessage(PID_BROWSER, msg);
end;

function TTrayIconHandler.Execute(const name: ustring; const obj: ICefv8Value;
  const arguments: TCefv8ValueArray; var retval: ICefv8Value;
  var exception: ustring): Boolean;

Var
  aLen,aID : Integer;

begin
  Result:=True;
  aLen:=length(arguments);
  Case name of
  SFuncAddMenuItem:
    begin
    if (aLen>1)
        and arguments[0].IsString
        and arguments[1].IsFunction then
      begin
      aID:=AddMenuItem(arguments[0].GetStringValue,arguments[1], TCefv8ContextRef.Current);
      retVal:=TCefv8ValueRef.NewInt(aID);
      end
    else
      exception:='Need 2 arguments: caption and callback';
    end;
  SFuncRemoveMenuItem:
    begin
    if (aLen>0) and arguments[0].IsInt then
      begin
      aID:=RemoveMenuItem(arguments[0].GetIntValue, TCefv8ContextRef.Current);
      retVal:=TCefv8ValueRef.NewInt(aID);
      end
    else
      exception:='Need 1 arguments: menu item ID';
    end;
  else
    Result:=False;
  end;
end;

{ TTrayMenu }

function TTrayMenu.AddMenu(const aCaption: uString): TTrayMenuItem;

begin
  Result:=Add as TTrayMenuItem;
  Result.Caption:=aCaption;
end;

Function TTrayMenu.RemoveMenu(const aID: Integer) : Boolean;

Var
  Itm : TCollectionItem;

begin
  Itm:=FindItemID(aID);
  Result:=Assigned(Itm);
  Itm.Free
end;

function TTrayMenu.GetM(AIndex: Integer): TTrayMenuItem;
begin
  Result:=Items[Aindex] as TTrayMenuItem;
end;

{ TLogHandler }



function TLogHandler.Execute(const name: ustring; const obj: ICefv8Value;
  const arguments: TCefv8ValueArray; var retval: ICefv8Value;
  var exception: ustring): Boolean;

Var
  msg: ICefProcessMessage;
  i : Integer;
  S : uString;

begin
  Result:=True;
  msg := TCefProcessMessageRef.New(SMsgDisplayLogMessage);
  msg.ArgumentList.SetSize(Length(Arguments));
  S:='';
  For I:=0 to Length(Arguments)-1 do
    begin
    if I>0 then
      S:=S+' ';
    S:=S+CefValToString(Arguments[i]);
    end;
  msg.ArgumentList.SetString(0,S );
  SendProcessMessage(PID_BROWSER, msg);
end;

end.
